/* kbdos.c -- MSDOS keyboard interrupt handler
 * Copyright (C) 1995, 1996 Markus F.X.J. Oberhumer
 * For conditions of distribution and use, see copyright notice in kb.h 
 */


#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <kb.h>
#include "_kb.h"


/***********************************************************************
// globals
************************************************************************/

int _kb_mode = 0;						/* do we have a keyboard handler? */
unsigned _kb_flags = 0;					/* flags from kb_install() */

volatile unsigned char _kb_key[128];	/* the key flags */
volatile unsigned short _kb_shift;		/* shift state */
volatile int _kb_keys_pressed;			/* number of keys pressed */


unsigned char _kb_prefix_scancode[128];
unsigned char _kb_inverse_prefix_scancode[128];
unsigned short _kb_shift_state_table[128];
unsigned short _kb_unshift_state_table[128];


/***********************************************************************
// static variables
// These are not volatile because all access outside the
//   interrupt handler uses KB_DISABLE().
************************************************************************/

#define KEY_BUFFER_SIZE 32			/* store up to 32 keys for kb_getch() */

static unsigned char _key_buffer[2*KEY_BUFFER_SIZE];
#define _key_buffer_start	(&_key_buffer[0])
#define _key_buffer_end		(&_key_buffer[sizeof(_key_buffer)])

static unsigned char *_key_buffer_head = _key_buffer_start;
static unsigned char *_key_buffer_tail = _key_buffer_start;

static unsigned char _t_prefix = 0;


/* debugging and error reporting */

#if defined(KB_DEBUG)
static unsigned char _kb_port_buffer[1024];
#define _kb_port_buffer_start	(&_kb_port_buffer[0])
#define _kb_port_buffer_end		(&_kb_port_buffer[sizeof(_kb_port_buffer)])
static unsigned char *_kb_port_buffer_head = _kb_port_buffer_start;

/* store unexpected situations in this error buffer */
static unsigned short _kb_error_buffer[1024];
#define _kb_error_buffer_start	(&_kb_error_buffer[0])
#define _kb_error_buffer_end	(&_kb_error_buffer[sizeof(_kb_error_buffer)])
static unsigned short *_kb_error_buffer_head = _kb_error_buffer_start;

#define KB_DEBUG_ERROR(code,scan) \
	if (_kb_error_buffer_head < _kb_error_buffer_end) \
		*_kb_error_buffer_head++ = (((code) << 8) | (scan))
#endif


/***********************************************************************
// reset all vars except _kb_flags
************************************************************************/

void _kb_clear(void)
{
	int i;

	KB_DISABLE();
	for (i = 0; i < 128; i++)
		_kb_key[i] = 0;
	_kb_shift = 0;
	_kb_keys_pressed = 0;

	_t_prefix = 0;

	_key_buffer_head = _key_buffer_tail = _key_buffer_start;
	KB_ENABLE();
}


/***********************************************************************
//
************************************************************************/

int kb_kbhit()
{
	register int c;

	if (!_kb_mode)
		return kbhit();

	KB_DISABLE();
	c = (_key_buffer_head != _key_buffer_tail);
	KB_ENABLE();

	return c;
}


/***********************************************************************
// basic keypress access.
// this function returns all relevant information:
//   the enhanced (mapped) scancode and the low bits of the shift status
// kb_getch() and kb_inkey() sit on top of this function.
************************************************************************/

unsigned kb_keypress(void)
{
	register unsigned char scan, shift;
	unsigned k;

	if (!_kb_mode)
		return 0;

	KB_DISABLE();
	if (_key_buffer_head == _key_buffer_tail)
	{
		KB_ENABLE();
		return 0;
	}

	scan = *_key_buffer_tail++;				/* get scan code */
	shift = *_key_buffer_tail++;			/* get shift code */
	if (_key_buffer_tail >= _key_buffer_end)
		_key_buffer_tail = _key_buffer_start;
	KB_ENABLE();

	k = scan | ((unsigned) shift << 8);
	return k;
}


/***********************************************************************
// int9 keyboard interrupt handler (IRQ1 keyboard data ready)
//   
// IMPORTANT NOTE: check your compiler flags so that no
//                 stack overflow checks will be inserted here !
//
// Most of this code should be quite obvious.
//   Prefixed keys (e0) are mapped to unique scancodes.
//
// The Pause key requires some extra handling:
//   I have observed the following:
//     Pressing Pause generates these 3 codes: e1 1d 45
//     Pressing Control+Pause (== Break) generates e0 46 which is
//       mapped to an unique virtual scancode.
//     Pressing both Control keys + Pause generates Pause and not e0 46.  
//     The release codes for Pause (e1 80+1d 80+45) and Control+Pause
//       (e0 80+46) are always generated after the press codes !
//       Releasing Pause or Control+Pause generates nothing - this is the
//       reason that there is no scancode for Pause, only the shift toggles.
//     The Pause key is the only key that generates e1.
//   Implementation:
//     The only thing to care for is the code 1d (KB_SCAN_LCONTROL)
//       in combination with the prefix count. 
//       This code is mapped to a special virtual key so that the
//       current shift status won't get touched.
//     Fortunately a plain e0 45 is never generated, so
//       the e1 1d 45 conversion works fine.
//   
// If this code looks confusing to you, just remove all KB_DEBUG stuff.
// I will remove it when this is out of beta stage.
************************************************************************/

#if defined(__BORLANDC__)
#pragma option -N-	// no check for stack overflow !
static void interrupt far _my_keyint()
{
	register unsigned char _t_scan, _t_tmp;
	register unsigned short _t_shift;

#elif defined(__EMX__)
static void _my_keyint()
{
	register unsigned char _t_scan, _t_tmp;
	register unsigned short _t_shift;

#elif defined(__GO32__)
	/* stop djgpp messing up with stack reads */
	static unsigned char _t_scan, _t_tmp;
	static unsigned short _t_shift;
static void _my_keyint(_go32_dpmi_registers * regs)
{

#elif defined(__WATCOMC__)
static void __interrupt __far _my_keyint()
{
	register unsigned char _t_scan, _t_tmp;
	register unsigned short _t_shift;

#else
#  error unsupported compiler
#endif

	_t_scan = KB_INP8(0x60);		/* read keyboard byte */

	/* tell the XT keyboard controller to clear the key */
	_t_tmp = KB_INP8(0x61);
	KB_OUTP8(0x61,_t_tmp | 0x80);
	KB_OUTP8(0x61,_t_tmp);

#if defined(KB_DEBUG)
	/* store port value in buffer */
	*_kb_port_buffer_head++ = _t_scan;	
	if (_kb_port_buffer_head >= _kb_port_buffer_end)
		_kb_port_buffer_head = _kb_port_buffer_start;
#endif

	if (_t_scan >= 0xe0)
	{
		if (_t_scan <= 0xe1)			/* 0xe0, 0xe1: special key prefix */
		{
			_t_prefix = _t_scan - (0xe0 - 1);	/* prefix count */
		}
		else if (_t_scan == 0xff)		/* handle overflow */
		{
			/* try it: press a lot of keys */
			_t_prefix = 0;
		}
		else						
		{
			/* shouldn't happen, but be paranoid */
			_t_prefix = 0;
#if defined(KB_DEBUG)
			KB_DEBUG_ERROR(0x01,_t_scan);
#endif
		}
	}
	else if (_t_scan & 0x80)			/* key was released */
	{
		_t_scan &= 0x7f;
#if defined(KB_DEBUG)
		if (_t_scan == 0 || _t_scan > KB_SCAN_F12)
			KB_DEBUG_ERROR(0x81,_t_scan);
#endif
		if (_t_prefix)			/* convert scancode */
		{
			--_t_prefix;
#if defined(KB_DEBUG)
			if (_t_scan == _kb_prefix_scancode[_t_scan])
				KB_DEBUG_ERROR(0x82,_t_scan);
			if (_t_prefix && _t_scan != KB_SCAN_LCONTROL)
				KB_DEBUG_ERROR(0x83,_t_scan);
#endif
			if (_t_prefix && _t_scan == KB_SCAN_LCONTROL)
				_t_scan = KB_SCAN_LCONTROL_VIRTUAL;
			else
				_t_scan = _kb_prefix_scancode[_t_scan];
		}

		if (_kb_key[_t_scan])
		{
			--_kb_keys_pressed;
			_kb_key[_t_scan] = 0;
			_kb_shift &= _kb_unshift_state_table[_t_scan];
		}
	}
	else								/* key was pressed */
	{
#if defined(KB_DEBUG)
		if (_t_scan == 0 || _t_scan > KB_SCAN_F12)
			KB_DEBUG_ERROR(0x01,_t_scan);
#endif
		if (_t_prefix)			/* convert scancode */
		{
			--_t_prefix;
#if defined(KB_DEBUG)
			if (_t_scan == _kb_prefix_scancode[_t_scan])
				KB_DEBUG_ERROR(0x02,_t_scan);
			if (_t_prefix && _t_scan != KB_SCAN_LCONTROL)
				KB_DEBUG_ERROR(0x03,_t_scan);
#endif
			if (_t_prefix && _t_scan == KB_SCAN_LCONTROL)
				_t_scan = KB_SCAN_LCONTROL_VIRTUAL;
			else
				_t_scan = _kb_prefix_scancode[_t_scan];
		}

		_t_shift = _kb_shift_state_table[_t_scan];

		if (!_kb_key[_t_scan])
		{
			if (_t_scan < KB_SCAN_VIRTUAL) /* do not count virtual keys */
			{
				++_kb_keys_pressed;
				_kb_key[_t_scan] = 1;
			}

			_kb_shift ^= _t_shift;	/* toggle shift status */
		}

		if (!(_t_shift & KB_SHIFT_VIRTUAL))		/* if not a shift key */
		{
			/* store keypress: scan code + low bits of shift state */
			*_key_buffer_head++ = _t_scan;
			*_key_buffer_head++ = (unsigned char) _kb_shift;

			/* adjust pointers */
			if (_key_buffer_head >= _key_buffer_end)
				_key_buffer_head = _key_buffer_start;
			if (_key_buffer_head == _key_buffer_tail)	/* buffer full */
			{
				_key_buffer_tail += 2;
				if (_key_buffer_tail >= _key_buffer_end)
					_key_buffer_tail = _key_buffer_start;
			}
		}
	}

	KB_OUTP8(0x20, 0x20);		/* ack. interrupt */
}


static void _my_keyint_end(void) { }



/***********************************************************************
// low level installation
************************************************************************/

#if defined(__BORLANDC__)

static void (interrupt far *_oldkeyint) (void) = NULL;

static int _kb_install(void)
{
	_oldkeyint = getvect(9);
	setvect(9, _my_keyint);
	return 1;
}

static void _kb_remove(void)
{
	if (_oldkeyint)
		setvect(9, _oldkeyint);
}

#endif /* __BORLANDC__ */



#if defined(__EMX__)
static int _kb_install(void)
{
	/* sorry, no interrupt handler possible with emx */
	return 0;
}
static void _kb_remove(void)
{
}
#endif /* __EMX__ */



#if defined(__GO32__)

/* This code is taken from the Allegro library written by Shawn Hargreaves.
 */

static _go32_dpmi_seginfo _key_rm_oldint;	  /* original real mode key IRQ */
static _go32_dpmi_seginfo _key_rm_int;	/* real mode interrupt segment info */
static _go32_dpmi_seginfo _key_pm_oldint;	  /* original prot-mode key IRQ */
static _go32_dpmi_seginfo _key_pm_int;	/* prot-mode interrupt segment info */
static _go32_dpmi_registers _key_rm_regs;

static int _kb_install(void)
{
	/* pm handler */
	_key_pm_int.pm_offset = (int) _my_keyint;
	_go32_dpmi_allocate_iret_wrapper(&_key_pm_int);
	_key_pm_int.pm_selector = _go32_my_cs();
	_go32_dpmi_get_protected_mode_interrupt_vector(9, &_key_pm_oldint);
	_go32_dpmi_set_protected_mode_interrupt_vector(9, &_key_pm_int);

	/* rm handler */
	memset(&_key_rm_regs, 0, sizeof(_go32_dpmi_registers));
	_key_rm_int.pm_offset = (int) _my_keyint;
	_go32_dpmi_allocate_real_mode_callback_iret(&_key_rm_int, &_key_rm_regs);
	_go32_dpmi_get_real_mode_interrupt_vector(9, &_key_rm_oldint);
	_go32_dpmi_set_real_mode_interrupt_vector(9, &_key_rm_int);

	return 1;
}

static void _kb_remove(void)
{
	_go32_dpmi_set_protected_mode_interrupt_vector(9, &_key_pm_oldint);
	_go32_dpmi_free_iret_wrapper(&_key_pm_int);
	_go32_dpmi_set_real_mode_interrupt_vector(9, &_key_rm_oldint);
	_go32_dpmi_free_real_mode_callback(&_key_rm_int);
}

#if defined(__DJGPP__)		/* djgpp v2 */

/* This locking code is based on work by Jonathan Paul Griffiths (JLib)
 * which itself is loosely based on work by Charles Sandman from the
 * DJGPP V2 beta 2 release of DJGPP. (file DPMIEXCP.C)
 * This software may be freely distributed, no warranty.
 */


static int _kb_lock_addr(void *addr, long size)
{
	u_short __kb_ds;
	__dpmi_meminfo kb_memblock;

	if (addr == NULL || size <= 0)
		return;

	/* get our ds  */
	__kb_ds = _go32_my_ds();

	__dpmi_get_segment_base_address(__kb_ds, &kb_memblock.address);
	kb_memblock.address += (unsigned) addr;
	kb_memblock.size = size;
	__dpmi_lock_linear_region(&kb_memblock);

	return 1;
}

#define KB_LOCK_CODE(a,b)	_kb_lock_addr((void *)(a),(char *)(b)-(char *)(a))
#define KB_LOCK_VAR(a)		_kb_lock_addr((void *)&(a),sizeof(a))

#endif /* __DJGPP__ */

#endif /* __GO32__ */



#if defined(__WATCOMC__)

static void (__interrupt __far *_oldkeyint) () = NULL;

static int _kb_install(void)
{
	_oldkeyint = _dos_getvect(9);
	_dos_setvect(9, _my_keyint);
	return 1;
}

static void _kb_remove(void)
{
	if (_oldkeyint)
		_dos_setvect(9, _oldkeyint);
}

#if defined(__386__)
static int _kb_lock_addr(void *addr, long size)
{
	/* FIXME: how do we lock in DOS/4GW ? */
	/* (we do not have to lock in PMODE/W) */

	return 1;
}

#define KB_LOCK_CODE(a,b)	_kb_lock_addr((void *)(a),(char *)(b)-(char *)(a))
#define KB_LOCK_VAR(a)		_kb_lock_addr((void *)&(a),sizeof(a))
#endif /* __386__ */

#endif /* __WATCOMC__ */



/***********************************************************************
// high level installation
************************************************************************/

int kb_install(unsigned flags)
{
	static unsigned char done = 0;
	unsigned bshift;
	unsigned short shift;

	_kb_flags = flags & 0xff;

	if (!done  &&  !(_kb_flags & KB_FLAG_NO_ATEXIT))
	{
		if (atexit(kb_remove) == 0)
		{
			done = 1;
			_kb_flags |= KB_FLAG_ATEXIT_INSTALLED;
		}
	}

	if (_kb_mode)
		return _kb_mode;

	/* get and convert current shift status */
	bshift = kb_bioskey(KB_KEYBRD_SHIFTSTATUS);
	shift = 0;
	if (bshift & 0x10)  shift |= KB_SHIFT_SCRLOCK;
	if (bshift & 0x20)  shift |= KB_SHIFT_NUMLOCK;
	if (bshift & 0x40)  shift |= KB_SHIFT_CAPSLOCK;
	if (bshift & 0x80)  shift |= KB_SHIFT_INSERT;

	_kb_init_tables();
	if (!_kb_lock())
		_kb_flags |= KB_FLAG_LOCK_FAILED;
	_kb_clear();
	if (_kb_flags & KB_FLAG_INIT_SHIFT_TOGGLES)
		_kb_shift = shift;
	KB_DISABLE();
	if (_kb_install())
		_kb_mode = 1;
	KB_ENABLE();

	return _kb_mode;
}


void kb_remove(void)
{
	if (!_kb_mode)
		return;

	KB_DISABLE();
	_kb_mode = 0;
	_kb_remove();
	KB_ENABLE();
	_kb_clear();
	_kb_unlock();
}



/***********************************************************************
// lock ANY addresses which may be seen by the keyboard interrupt
************************************************************************/

static int _kb_locked = 0;

int _kb_lock(void)
{
	int l = 1;

	if (_kb_locked)
		return 1;

#if defined(KB_LOCK_CODE)
	/* interrupt handler */
	l &= KB_LOCK_CODE(_my_keyint,_my_keyint_end);
#endif

#if defined(KB_LOCK_VAR)
	/* public globals */
	l &= KB_LOCK_VAR(_kb_key);
	l &= KB_LOCK_VAR(_kb_shift);
	l &= KB_LOCK_VAR(_kb_keys_pressed);

	/* internal globals */
	l &= KB_LOCK_VAR(_kb_prefix_scancode);
	l &= KB_LOCK_VAR(_kb_shift_state_table);
	l &= KB_LOCK_VAR(_kb_unshift_state_table);

	/* statics */
	l &= KB_LOCK_VAR(_key_buffer);
	l &= KB_LOCK_VAR(_key_buffer_head);
	l &= KB_LOCK_VAR(_key_buffer_tail);
	l &= KB_LOCK_VAR(_t_prefix);
#if defined(__GO32__)
	l &= KB_LOCK_VAR(_t_scan);
	l &= KB_LOCK_VAR(_t_tmp);
#endif

#if defined(KB_DEBUG)
	l &= KB_LOCK_VAR(_kb_port_buffer);
	l &= KB_LOCK_VAR(_kb_port_buffer_head);
	l &= KB_LOCK_VAR(_kb_error_buffer);
	l &= KB_LOCK_VAR(_kb_error_buffer_head);
#endif
#endif

	if (l)
		_kb_locked = 1;
	return l;
}


void _kb_unlock(void)
{
	if (!_kb_locked)
		return;

	/* FIXME - do nothing for now */
	/* _kb_locked = 0; */
}


/***********************************************************************
// debugging and error reporting
************************************************************************/

void _kb_port_debug(void)
{
#if defined(KB_DEBUG)
	static unsigned prefix = 0;
	unsigned char *p;
	/* we'd better go volatile here */
	unsigned char ** volatile bh = &_kb_port_buffer_head;

	if (!_kb_mode)
		return;

	for (p = _kb_port_buffer_start; p < *bh; p++)
	{
		unsigned c = *p;

		if (c == 0xe0 || c == 0xe1)
			prefix = c - (0xe0 - 1);
		else
		{
			if (prefix)
				printf("%02x+", prefix + (0xe0 - 1));

			if (c < 0xe0 && (c & 0x80))
				printf("80+%02x ", c & 0x7f);
			else
				printf("%02x ", c);

			if (prefix)
				prefix--;
		}
	}

	/* reset buffer pointer */
	_kb_port_buffer_head = _kb_port_buffer_start;
#endif
}


int _kb_error_debug_avail(void)
{
#if defined(KB_DEBUG)
	return _kb_error_buffer_head != _kb_error_buffer_start;
#else
	return 0;
#endif
}

void _kb_error_debug(void)
{
#if defined(KB_DEBUG)
	unsigned short *p;
	/* we'd better go volatile here */
	unsigned short ** volatile bh = &_kb_error_buffer_head;

	for (p = _kb_error_buffer_start; p < *bh; p++)
		printf("%04x ", (unsigned) *p);

	/* reset buffer pointer */
	_kb_error_buffer_head = _kb_error_buffer_start;
#endif
}


/*
vi:ts=4
*/
